/* **************************************************************************
 * Copyright (C) 2019-2025 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * *************************************************************************/

#include "roclapack_geqrf.hpp"

ROCSOLVER_BEGIN_NAMESPACE

template <typename T, typename U>
rocblas_status rocsolver_geqrf_batched_impl(rocblas_handle handle,
                                            const rocblas_int m,
                                            const rocblas_int n,
                                            U A,
                                            const rocblas_int lda,
                                            T* ipiv,
                                            const rocblas_stride stridep,
                                            const rocblas_int batch_count)
{
    ROCSOLVER_ENTER_TOP("geqrf_batched", "-m", m, "-n", n, "--lda", lda, "--strideP", stridep,
                        "--batch_count", batch_count);

    if(!handle)
        return rocblas_status_invalid_handle;

    // argument checking
    rocblas_status st = rocsolver_geqr2_geqrf_argCheck(handle, m, n, lda, A, ipiv, batch_count);
    if(st != rocblas_status_continue)
        return st;

    // working with unshifted arrays
    rocblas_int shiftA = 0;

    // batched execution
    rocblas_stride strideA = 0;

    // memory workspace sizes:
    // size for constants in rocblas calls
    size_t size_scalars;
    // size of arrays of pointers (for batched cases) and re-usable workspace
    bool optim_mem;
    size_t size_work_workArr_work1, size_work2, size_work3, size_work4, size_workArr;
    // extra requirements for calling GEQR2 and to store temporary triangular factor
    size_t size_Abyx_norms_trfact;
    // extra requirements for calling GEQR2 and LARFB
    size_t size_diag_tmptr;
    rocsolver_geqrf_getMemorySize<true, false, T>(
        m, n, batch_count, &size_scalars, &size_work_workArr_work1, &size_work2, &size_work3,
        &size_work4, &size_Abyx_norms_trfact, &size_diag_tmptr, &size_workArr, &optim_mem);

    if(rocblas_is_device_memory_size_query(handle))
        return rocblas_set_optimal_device_memory_size(
            handle, size_scalars, size_work_workArr_work1, size_work2, size_work3, size_work4,
            size_Abyx_norms_trfact, size_diag_tmptr, size_workArr);

    // memory workspace allocation
    void *scalars, *work_workArr_work1, *work2, *work3, *work4, *Abyx_norms_trfact, *diag_tmptr,
        *workArr;
    rocblas_device_malloc mem(handle, size_scalars, size_work_workArr_work1, size_work2, size_work3,
                              size_work4, size_Abyx_norms_trfact, size_diag_tmptr, size_workArr);

    if(!mem)
        return rocblas_status_memory_error;

    scalars = mem[0];
    work_workArr_work1 = mem[1];
    work2 = mem[2];
    work3 = mem[3];
    work4 = mem[4];
    Abyx_norms_trfact = mem[5];
    diag_tmptr = mem[6];
    workArr = mem[7];
    if(size_scalars > 0)
        init_scalars(handle, (T*)scalars);

    // execution
    return rocsolver_geqrf_template<true, false, T>(
        handle, m, n, A, shiftA, lda, strideA, ipiv, stridep, batch_count, (T*)scalars,
        work_workArr_work1, work2, work3, work4, (T*)Abyx_norms_trfact, (T*)diag_tmptr,
        (T**)workArr, optim_mem);
}

ROCSOLVER_END_NAMESPACE

/*
 * ===========================================================================
 *    C wrapper
 * ===========================================================================
 */

extern "C" {

rocblas_status rocsolver_sgeqrf_batched(rocblas_handle handle,
                                        const rocblas_int m,
                                        const rocblas_int n,
                                        float* const A[],
                                        const rocblas_int lda,
                                        float* ipiv,
                                        const rocblas_stride stridep,
                                        const rocblas_int batch_count)
{
    return rocsolver::rocsolver_geqrf_batched_impl<float>(handle, m, n, A, lda, ipiv, stridep,
                                                          batch_count);
}

rocblas_status rocsolver_dgeqrf_batched(rocblas_handle handle,
                                        const rocblas_int m,
                                        const rocblas_int n,
                                        double* const A[],
                                        const rocblas_int lda,
                                        double* ipiv,
                                        const rocblas_stride stridep,
                                        const rocblas_int batch_count)
{
    return rocsolver::rocsolver_geqrf_batched_impl<double>(handle, m, n, A, lda, ipiv, stridep,
                                                           batch_count);
}

rocblas_status rocsolver_cgeqrf_batched(rocblas_handle handle,
                                        const rocblas_int m,
                                        const rocblas_int n,
                                        rocblas_float_complex* const A[],
                                        const rocblas_int lda,
                                        rocblas_float_complex* ipiv,
                                        const rocblas_stride stridep,
                                        const rocblas_int batch_count)
{
    return rocsolver::rocsolver_geqrf_batched_impl<rocblas_float_complex>(handle, m, n, A, lda, ipiv,
                                                                          stridep, batch_count);
}

rocblas_status rocsolver_zgeqrf_batched(rocblas_handle handle,
                                        const rocblas_int m,
                                        const rocblas_int n,
                                        rocblas_double_complex* const A[],
                                        const rocblas_int lda,
                                        rocblas_double_complex* ipiv,
                                        const rocblas_stride stridep,
                                        const rocblas_int batch_count)
{
    return rocsolver::rocsolver_geqrf_batched_impl<rocblas_double_complex>(
        handle, m, n, A, lda, ipiv, stridep, batch_count);
}

rocblas_status rocsolver_sgeqrf_batched_64(rocblas_handle handle,
                                           const int64_t m,
                                           const int64_t n,
                                           float* const A[],
                                           const int64_t lda,
                                           float* ipiv,
                                           const rocblas_stride stridep,
                                           const int64_t batch_count)
{
#ifdef HAVE_ROCBLAS_64
    return rocsolver::rocsolver_geqrf_batched_impl<float>(handle, m, n, A, lda, ipiv, stridep,
                                                          batch_count);
#else
    return rocblas_status_not_implemented;
#endif
}

rocblas_status rocsolver_dgeqrf_batched_64(rocblas_handle handle,
                                           const int64_t m,
                                           const int64_t n,
                                           double* const A[],
                                           const int64_t lda,
                                           double* ipiv,
                                           const rocblas_stride stridep,
                                           const int64_t batch_count)
{
#ifdef HAVE_ROCBLAS_64
    return rocsolver::rocsolver_geqrf_batched_impl<double>(handle, m, n, A, lda, ipiv, stridep,
                                                           batch_count);
#else
    return rocblas_status_not_implemented;
#endif
}

rocblas_status rocsolver_cgeqrf_batched_64(rocblas_handle handle,
                                           const int64_t m,
                                           const int64_t n,
                                           rocblas_float_complex* const A[],
                                           const int64_t lda,
                                           rocblas_float_complex* ipiv,
                                           const rocblas_stride stridep,
                                           const int64_t batch_count)
{
#ifdef HAVE_ROCBLAS_64
    return rocsolver::rocsolver_geqrf_batched_impl<rocblas_float_complex>(handle, m, n, A, lda, ipiv,
                                                                          stridep, batch_count);
#else
    return rocblas_status_not_implemented;
#endif
}

rocblas_status rocsolver_zgeqrf_batched_64(rocblas_handle handle,
                                           const int64_t m,
                                           const int64_t n,
                                           rocblas_double_complex* const A[],
                                           const int64_t lda,
                                           rocblas_double_complex* ipiv,
                                           const rocblas_stride stridep,
                                           const int64_t batch_count)
{
#ifdef HAVE_ROCBLAS_64
    return rocsolver::rocsolver_geqrf_batched_impl<rocblas_double_complex>(
        handle, m, n, A, lda, ipiv, stridep, batch_count);
#else
    return rocblas_status_not_implemented;
#endif
}

} // extern C
